//
//  mtonesdk.swift
//  BuyersHelper
//
//  Created by John McKnight on 7/21/14.
//  Copyright (c) 2014 Media Tuners LLC. All rights reserved.
//
//  VERSION: 1.0.0.3
//    Changed GPS handling, added a flag to suppress diagnostic alerts and moved the test to the code that starts bluetooth, changed proximity_get.  Added applicationDidEnterBackground and applicationDidBecomeActive to SDK, these functions MUST be called from the AppDelegate to ensure the app is battery friendly. Added autoConfigure flag.
//  VERSION: 1.0.0.2
//    Suppressed diagnostic for Homebrite since it isn't using beacons.
//  VERSION: 1.0.0.1
//    Changed logging and filtering of beacons that fail to return useful beacon information 5 times in a row.  If the beacon can't return data, it will not be checked until 24 hours later.

import Foundation
import UIKit
import CoreLocation
import CoreBluetooth
import SystemConfiguration
import AddressBook

// MTOne SDK support.
// Needs Network Access, BlueTooth, GPS, JSON decoding, calls to standard endpoints


@objc public class Serializable : NSObject{
    func toDictionary() -> NSDictionary {
        var aClass : AnyClass? = self.dynamicType
        var propertiesCount : CUnsignedInt = 0
        let propertiesInAClass : UnsafeMutablePointer<objc_property_t> = class_copyPropertyList(aClass, &propertiesCount)
        var propertiesDictionary : NSMutableDictionary = NSMutableDictionary()
        
        for var i = 0; i < Int(propertiesCount); i++ {
            var property = propertiesInAClass[i]
            var propName = NSString(CString: property_getName(property), encoding: NSUTF8StringEncoding)
            var propType = property_getAttributes(property)
            var propValue : AnyObject! = self.valueForKey(propName!);
            
            if propValue is Serializable {
                propertiesDictionary.setValue((propValue as Serializable).toDictionary(), forKey: propName!)
            } else if propValue is Array<Serializable> {
                var subArray = Array<NSDictionary>()
                for item in (propValue as Array<Serializable>) {
                    subArray.append(item.toDictionary())
                }
                propertiesDictionary.setValue(subArray, forKey: propName!)
            } else if propValue is NSData {
                propertiesDictionary.setValue((propValue as NSData).base64EncodedStringWithOptions(nil), forKey: propName!)
            } else if propValue is Double {
                propertiesDictionary.setValue((propValue as Double), forKey: propName!)
            } else if propValue is Float {
                propertiesDictionary.setValue((propValue as Float), forKey: propName!)
            }  else if propValue is Int {
                propertiesDictionary.setValue((propValue as Int), forKey: propName!)
            } else if propValue is Bool {
                propertiesDictionary.setValue((propValue as Bool).boolValue, forKey: propName!)
            } else if propValue is NSDate {
                var date = propValue as NSDate
                let dateFormatter = NSDateFormatter()
                dateFormatter.dateFormat = "Z"
                var dateString = NSString(format: "/Date(%.0f000%@)/", date.timeIntervalSince1970, dateFormatter.stringFromDate(date))
                propertiesDictionary.setValue(dateString, forKey: propName!)
            } else {
                propertiesDictionary.setValue(propValue, forKey: propName!)
            }
        }
        
        return propertiesDictionary
    }
    
    func toJson() -> NSData! {
        var dictionary = self.toDictionary()
        var err: NSError?
        return NSJSONSerialization.dataWithJSONObject(dictionary, options:NSJSONWritingOptions(0), error: &err)
    }
    
    func toJsonString() -> NSString! {
        return NSString(data: self.toJson(), encoding: NSUTF8StringEncoding)
    }
    
    override init() {
    }
}


@objc public class ServerResponse {
    final var json: JSON? = nil
    var status: Int? = nil
    var error:NSError? = nil
    var success:Bool = false
    final var data:JSON?
    final var action:JSON?
    
    init(success:Bool, json: JSON, status:Int, error:NSError? = nil) {
        self.success = success
        self.json = json
        self.status = status
        self.error = error
        self.data = nil
        self.action = nil
    }
    
    public func errorMessageAsString() -> String {
        if let message = json!["response"]["information"]["status"].asString {
            return message
        } else {
            return ""
        }
    }
    
    public func showError() {
        let error = self.errorMessageAsString()
        
        if error != "" {
            MTOneSharedInstance.errorAlert(error)
        }
    }
}


public enum BeaconFilter {
    case None
    case KnownOnly
    case KnownOnlyMatchesProximity
}


@objc public class MTOneLogBeacon:Serializable {
    var beacon_id:String = ""
    var zone:String = ""
}


@objc public class MTOneProximityLogEntry:Serializable {
    var client_event_timestamp:NSTimeInterval = NSDate().timeIntervalSince1970
    var dwell_time:Double = 0
    
    var beacon_id:String = ""
    var beacon_zone:String = ""
    var location_id:String = ""
    var place_id:String = ""
    var handler_count:Int = 0
    var lat:Double = 0
    var lng:Double = 0
    var other_beacons_in_range:[MTOneLogBeacon?] = []
    
    var sent_to_server:Bool = false
    
    public func debug() {
        ("beacon_id: \(beacon_id), beacon_zone: \(beacon_zone), location_id: \(location_id), place_id: \(place_id), handler_count: \(handler_count), lat: \(lat), lng: \(lng), dwell_time: \(dwell_time)")
    }
}


@objc public class MTOneLogger:NSObject, UIApplicationDelegate {
    var logEntries:[MTOneProximityLogEntry?] = []
    var isSending:Bool = false
    var lastSend:Double = 0
    var timer:NSTimer? = nil
    var lastTimerCall:Double = NSDate().timeIntervalSince1970
    
    class var sharedInstance:MTOneLogger {
        return MTOneSharedLogger
    }
    
    
    override init() {
        var timeSinceLastSignal:NSDate? = nil
        
        super.init()
    }
    
    
    internal func restartTimerIfTooLong() {
        if self.lastTimerCall == 0 {
            return
        }
        
        let now = NSDate().timeIntervalSince1970
        
        if (now - self.lastTimerCall) > 90 {
            if self.timer != nil {
                self.timer!.invalidate()
                self.timer = nil
            }
            
            self.lastTimerCall = 0
            
            startTimer()
        }
    }
    
    internal func startTimer() {
        if self.timer == nil {
            self.timer = NSTimer.scheduledTimerWithTimeInterval(60,
                target: self,
                selector: Selector("tick"),
                userInfo: nil,
                repeats: true)
        }
    }
    
    
    public func clear() {
        isSending = false
        logEntries = []
        lastSend = 0
    }
    
    
    @objc func tick() {
        self.restartTimerIfTooLong()
        
        self.lastTimerCall = NSDate().timeIntervalSince1970
        
        let now = NSDate().timeIntervalSince1970
        
        if (now - self.lastSend) > 60 {
            if logEntries.isEmpty == true {
                return
            }
            
            self.sendEntriesToServer(force: false, n: logEntries.count)
        }
    }
    
    
    public func debug() {
        if logEntries.isEmpty {
            println("No Entries in Beacon Log")
            return
        }
        
        println("BEACON LOG")
        
        for i in 0..<logEntries.count {
            println("Beacon Entry: \(i)")
            println(logEntries[i]?.debug())
        }
        
    }
    
    
    // This function sets the dwell of the last entry.  Called by no beacon or a beacon change
    public func markEnd() {
        if logEntries.isEmpty {
            return
        }
        
        var logEntry:MTOneProximityLogEntry? = logEntries.last!
        
        if logEntry != nil {
            let now:NSTimeInterval = NSDate().timeIntervalSince1970
            if let then:NSTimeInterval = logEntry?.client_event_timestamp {
                // Only set the dwell if it hasn't already been done
                if logEntry?.dwell_time == 0 {
                    logEntry?.dwell_time = now - then
                }
            }
            
            self.lastSend = NSDate().timeIntervalSince1970
            self.sendEntriesToServer(force: false)
        }
    }
    
    
    // Get n Entries which have NOT been sent to the server
    public func getNEntries(n: Int) -> [MTOneProximityLogEntry?] {
        var newEntries:[MTOneProximityLogEntry?] = []
        
        if logEntries.isEmpty {
            return newEntries
        }
        
        for i in 0..<logEntries.count {
            var logEntry:MTOneProximityLogEntry? = logEntries[i]
            
            // It is only valid if it hasn't been sent and has a dwell_time that is not 0
            if logEntry?.sent_to_server == false && logEntry?.dwell_time > 0 {
                newEntries.append(logEntry)
                
                if newEntries.count >= n {
                    return newEntries
                }
            }
        }
        
        return newEntries
    }
    
    
    // Consider adding a completion handler
    public func sendEntriesToServer(force: Bool = false, n: Int = 5) {
        self.restartTimerIfTooLong()
        
        if isSending == true {
            return
        }
        
        var entriesToSend = self.getNEntries(n)
        
        // Exit if there were none
        if entriesToSend.isEmpty {
            return
        }
        
        
        // Exit if it is less than we wanted to send
        if force == false {
            if entriesToSend.count < n {
                return
            }
        }
        
        
        if MTOneSharedInstance.isConnectedToNetwork() == false {
            return
        }
        
        isSending = true
        
        var json:String = "["
        for i in 0..<entriesToSend.count {
            var logEntry:MTOneProximityLogEntry? = logEntries[i]
            json += logEntry!.toJsonString()
            
            if i < (entriesToSend.count - 1) {
                json += ","
            }
            
        }
        
        json += "]";
        
        if MTOneSharedInstance.token != nil {
            // Send the records to the server
            var token:String = MTOneSharedInstance.token!
            
            let parms = ["service": "all", "token": token, "type": "proximity", "log": json] as Dictionary<String, String>
            
            MTOneSharedInstance.callMTOne("user/log", params: parms, completion: { (response) -> Void in
                if response.status == 200 {
                    self.purgeSentItems(entriesToSend);
                } else {
                }
                self.isSending = false
            })
        }
    }
    
    
    private func purgeSentItems(entries: [MTOneProximityLogEntry?]) {
        if entries.isEmpty {
            return
        }
        
        // We need to remove the item from the main list
        for i in 0..<entries.count {
            let logEntry:MTOneProximityLogEntry? = entries[i]
            for j in 0..<logEntries.count {
                if logEntry == logEntries[j] {
                    logEntries.removeAtIndex(j)
                    break
                }
            }
        }
    }
    
    
    // This sends the first item to the server
    public func sendFirstEntryToServer() {
        if logEntries.isEmpty {
            return
        }
        
        var logEntry:MTOneProximityLogEntry? = logEntries.first!
        
        if logEntry != nil {
            var sJson:String = logEntry!.toJsonString()
            var token:String = MTOneSharedInstance.token!
            
            let parms = ["service": "all", "token": token, "type": "proximity", "log": sJson] as Dictionary<String, String>
            
            MTOneSharedInstance.callMTOne("user/log", params: parms, completion: { (response) -> Void in
                NSLog("Log write call completed");
            })
        }
    }
    
    
    // This sends the first item to the server
    public func sendLastEntryToServer() {
        if logEntries.isEmpty {
            return
        }
        
        var logEntry:MTOneProximityLogEntry? = logEntries.last!
        logEntries.removeLast()
        
        if logEntry != nil {
            var sJson:String = logEntry!.toJsonString()
            var token:String = MTOneSharedInstance.token!
            
            let parms = ["service": "all", "token": token, "type": "proximity", "log": sJson] as Dictionary<String, String>
            
            MTOneSharedInstance.callMTOne("user/log", params: parms, completion: { (response) -> Void in
                NSLog("Log write call completed");
            })
        }
    }
    
    public func noBeacon() {
        self.markEnd()
    }
    
    
    public func addBeaconEntry(location: CLLocation?, beacon_id:String, beacon_zone: String, location_id: String, place_id: String, handler_count: Int, other_beacons:[CLBeacon?]) {
        
        self.markEnd()
        
        var newEntry:MTOneProximityLogEntry = MTOneProximityLogEntry()
        
        //newEntry.client_event_timestamp = NSDate().timeIntervalSince1970
        newEntry.beacon_id = beacon_id
        newEntry.beacon_zone = beacon_zone
        newEntry.location_id = location_id
        newEntry.place_id = place_id
        newEntry.handler_count = handler_count
        
        // set lat, lng if available
        if location != nil {
            newEntry.lat = Double(location!.coordinate.latitude)
            newEntry.lng = Double(location!.coordinate.longitude)
        }
        
        // Assemble the other beacons from the beacon array
        if other_beacons.isEmpty == false {
            for i in 0..<other_beacons.count {
                var newBeacon = MTOneLogBeacon()
                newBeacon.beacon_id = MTOneSharedInstance.makeFullUUID(other_beacons[i])
                newBeacon.zone = MTOneSharedInstance.getBeaconProximityAsString(other_beacons[i]!)
                
                newEntry.other_beacons_in_range.append(newBeacon)
            }
        }
        
        
        logEntries.append(newEntry)
        
        
        if MTOneSharedInstance.isRunningInBackground() {
            self.sendEntriesToServer(force: true)
        }
    }
}


@objc public class MTOneLinkAction:Serializable {
    var handler:String = "far"
    var uri:String = ""
    var callback_uri:String = ""
    var label:String = ""
    var icon_uri:String = ""
    var mime_type:String = "text/html"
    var attachments:[MTOneAttachment] = []
}


@objc public class MTOneProximityPlace:Serializable {
    var place_id: String = ""
    var values:Dictionary<String, String> = [:]
    
    public func addValue(key: String, value: String) {
        self.values[key] = value
    }
}


@objc public class MTOneProximityBeacon:Serializable {
    var type:String = "ibeacon"
    var accuracy:Double = 0
    var beacon_id:String = ""
    var ibeacon_uuid:String = ""
    var beacon_major:String = ""
    var beacon_minor:String = ""
    
    public func setBeacon(ibeacon_uuid: String, beacon_major: String, beacon_minor: String) {
        self.ibeacon_uuid = ibeacon_uuid
        self.beacon_major = beacon_major
        self.beacon_minor = beacon_minor
        self.beacon_id = "\(ibeacon_uuid)-\(beacon_major)-\(beacon_minor)"
    }
    
    
    // Explodes the beacon ID into its components
    public func setBeaconId(beacon_id: String) {
        self.beacon_id = beacon_id
        
        var beacon_bits = beacon_id.componentsSeparatedByString("_")
        if beacon_bits.count == 3 {
            self.ibeacon_uuid = beacon_bits[0]
            self.beacon_major = beacon_bits[1]
            self.beacon_minor = beacon_bits[2]
        } else {
            NSLog("Invalid beacon_id: " + beacon_id)
        }
    }
}


@objc public class MTOneProximityLocation:Serializable {
    var location_id:String = ""
    var key:String = ""
    var icon_uri = ""
    var values:Dictionary<String, String> = [:]
    var actions:[AnyObject] = []
    var point:MTOnePoint = MTOnePoint()
    var _description:String = ""
    var name:String = ""
    
    public func addValue(key: String, value: String) {
        self.values[key] = value
    }
}

/*
{
"rating": 0,
"service_item_id": "yelp_coffee",
"description": "",
"point": null,
"global_item_id": "b055aa463ac5ff831adfe233bc0a0475",
"actions": [
{
"handler": "list",
"uri": "https:\\/\\/mtone-dev.tuner2.net\\/api\\/v3\\/poi\\/search\\/?service=yelp&search=coffee",
"callback_uri": ""
}
],
"service_icon_uri": "http:\\/\\/mtone-dev.tuner2.net\\/api\\/v3\\/_assets\\/service_yelp.png",
"service_action_uri": "",
"list_priority": 5,
"info": {},
"name": "Coffee & Tea",
"service": "yelp",
"icon_uri": null,
"type": "poi",
"display": "list"
}
*/


@objc public class MTOneActions:Serializable {
    var callback:String = ""
    var handler:String = ""
    var uri:String = ""
    
    override init() {
        super.init()
    }
    
    
    // This creates the dictionary that is ued to set values
    override func toJsonString() -> NSString! {
        var newDict:Dictionary<String, AnyObject> = [:]
        
        return MTOneSharedInstance.JSONStringify(newDict)
    }
}


@objc public class MTOneAction:Serializable {
    var actions:[MTOneActions] = []
    var _description:String = ""
    var display:String = "list"
    var global_item_id:String = ""
    var icon_uri:String = ""
    var info:MTOneInfo = MTOneInfo()
    var list_priority:String = ""
    var name:String = ""
    var point:MTOnePoint = MTOnePoint()
    var rating: Int = 0
    var service:String = "mtone"
    var service_action_uri:String = ""
    var service_icon_uri:String = ""
    var service_item_id:String = ""
    var type:String = "poi"
    
    override init() {
        super.init()
    }
    
    
    // This creates the dictionary that is ued to set values
    override func toJsonString() -> NSString! {
        var newDict:Dictionary<String, AnyObject> = [:]
        
        // Convert the array of actions into a JSON representation
        var joiner:String = ","
        var newArray:[String] = []
        for action:MTOneActions in actions {
            newArray.append(action.toJsonString())
        }
        newDict["actions"] = joiner.join(newArray)
        
        
        if _description != "" {
            newDict["description"] = description
        }
        
        if display != "" {
            newDict["display"] = display
        }
        
        if global_item_id != "" {
            newDict["global_item_id"] = global_item_id
        }
        
        if icon_uri != "" {
            newDict["icon_uri"] = icon_uri
        }
        
        newDict["info"] = info.toJsonString()
        
        if name != "" {
            newDict["name"] = name
        }
        
        newDict["point"] = point.toJsonString()
        
        newDict["rating"] = rating
        
        if service != "" {
            newDict["service"] = service
        }
        
        if service_action_uri != "" {
            newDict["service_action_uri"] = service_action_uri
        }
        
        if service_icon_uri != "" {
            newDict["service_icon_uri"] = service_icon_uri
        }
        
        if service_item_id != "" {
            newDict["service_item_id"] = service_item_id
        }
        
        if type != "" {
            newDict["type"] = type
        }
        
        return MTOneSharedInstance.JSONStringify(newDict)
    }
}


@objc public class MTOneFavoriteAction:MTOneAction {
    override init() {
        super.init()
    }
    
    
    // This creates the dictionary that is ued to set values
    override func toJsonString() -> NSString! {
        var newDict:Dictionary<String, AnyObject> = [:]
        
        return MTOneSharedInstance.JSONStringify(newDict)
    }
}


@objc public class MTOneAttachment:Serializable {
    var name: String = ""
    var data: String = ""
    var mime_type: String = ""
    
    init(_ name : String, mime_type: String, data: AnyObject, no_image_resize: Bool) {
        super.init()
        
        self.name = name
        self.mime_type = mime_type
        
        if self.mime_type == "image/jpg" {
            self.mime_type = "image/jpeg"
        }
        
        var text_data: String = ""
        
        // Resize the image and store it
        if self.mime_type == "image/jpeg" || self.mime_type == "image/png" {
            if let check_image: UIImage = data as? UIImage {
                var image:UIImage
                
                if no_image_resize == false {
                    image = UIImage.smartResize(check_image, width: 2048, height: 1536)
                } else {
                    image = check_image
                }
                
                var image_data:NSData = UIImageJPEGRepresentation(image, 0.8);
                
                //self.data = image_data.hexString()
                self.data = image_data.base64EncodedStringWithOptions(NSDataBase64EncodingOptions.Encoding64CharacterLineLength) as String
                self.data = self.data.stringByAddingPercentEncodingWithAllowedCharacters(.URLHostAllowedCharacterSet())!
            }
        } else {
            self.data = data.base64EncodedStringWithOptions(NSDataBase64EncodingOptions.Encoding64CharacterLineLength) as String
        }
    }
}


@objc public class MTOneDictionary {
    var dictionary:Dictionary<String, AnyObject> = [:]
    var nativeType:Dictionary<String, String> = [:]
    
    init() {
        self.clear()
    }
    
    
    public func clear() {
        self.dictionary = [:]
        self.nativeType = [:]
    }
    
    
    public func count() -> Int {
        return dictionary.count
    }
    
    
    public func remove(key: String) {
        self.dictionary.removeValueForKey(key)
        self.nativeType.removeValueForKey(key)
    }
    
    
    public func has(key: String) -> Bool {
        return self.dictionary[key] != nil
    }
    
    
    public func addString(key:String, value: String) {
        self.nativeType[key] = "String"
        self.dictionary[key] = NSString(string: value)
    }
    
    
    public func addDouble(key:String, value:Double) {
        self.nativeType[key] = "Double"
        self.dictionary[key] = NSNumber(double: value)
    }
    
    
    public func addFloat(key:String, value:Float) {
        self.nativeType[key] = "Float"
        self.dictionary[key] = NSNumber(float: value)
    }
    
    
    public func addInt(key:String, value:Int) {
        self.nativeType[key] = "Int"
        self.dictionary[key] = NSNumber(integer: value)
    }
    
    
    public func addBool(key:String, value:Bool) {
        self.nativeType[key] = "Bool"
        self.dictionary[key] = NSNumber(bool: value)
    }
    
    
    public func get(key:String) -> AnyObject? {
        if self.has(key) {
            return self.dictionary[key]
        } else {
            return nil
        }
    }
    
    
    // Retrieves a value as a type
    public func asString(key:String) -> String {
        if self.has(key) {
            return self.dictionary[key] as String
        } else {
            return ""
        }
    }
    
    
    public func asDouble(key:String) -> Double {
        if self.has(key) {
            let value:AnyObject? = self.dictionary[key]
            
            if value!.isKindOfClass(NSNumber) {
                let nValue:NSNumber = value! as NSNumber
                return nValue.doubleValue
            } else {
                return Double(0)
            }
        } else {
            return Double(0)
        }
    }
    
    
    public func asFloat(key:String) -> Float {
        if self.has(key) {
            let value:AnyObject? = self.dictionary[key]
            
            if value!.isKindOfClass(NSNumber) {
                let nValue:NSNumber = value! as NSNumber
                return nValue.floatValue
            } else {
                return Float(0)
            }
        } else {
            return Float(0)
        }
    }
    
    
    public func asInt(key:String) -> Int {
        if self.has(key) {
            let value:AnyObject? = self.dictionary[key]
            
            if value!.isKindOfClass(NSNumber) {
                let nValue:NSNumber = value! as NSNumber
                return nValue.integerValue
            } else {
                return Int(0)
            }
        } else {
            return Int(0)
        }
    }
    
    
    public func asBool(key:String) -> Bool {
        if self.has(key) {
            let value:AnyObject? = self.dictionary[key]
            
            if value!.isKindOfClass(NSNumber) {
                let nValue:NSNumber = value! as NSNumber
                return nValue.boolValue
            } else {
                return false
            }
        } else {
            return false
        }
    }
    
    
    public func each (eachFunction each: (key:String, value:AnyObject) -> ()) {
        for (key, value) in self.dictionary {
            each(key: key, value: value)
        }
    }
    
    
    public func fromJsonDictionary(sourceDict:Dictionary<String, AnyObject>) {
        self.clear()
        
        for (key, value) in sourceDict {
            if let j = value as? JSON {
                let n:Double? = j.asDouble
                if n != nil {
                    self.addString(key, value: String(format:"%f", n!))
                }
                
                let s:String? = j.asString
                if s != nil {
                    self.addString(key, value: s!)
                }
            }
        }
    }
    
    
    public func nativeType(key: String) -> String {
        var result = "Unknown"
        
        if self.has(key) {
            result = self.nativeType[key]!
        }
        
        return result
    }
}


@objc public class MTOnePoint:Serializable {
    public var lat:Double = 0.0
    public var lng:Double = 0.0
}


@objc public class MTOneBeacon:Serializable {
    public var point:MTOnePoint = MTOnePoint()
    public var accuracy:Double = 0.0
    
    public func fromJson(beacon:JSON?) {
        if beacon != nil {
            if let accuracy = beacon!["accuracy"].asDouble {
                self.accuracy = accuracy
            }
            
            if let lat = beacon!["point"]["lat"].asDouble {
                self.point.lat = lat
            }
            
            if let lng = beacon!["point"]["lng"].asDouble {
                self.point.lng = lng
            }
        }
    }
}


@objc public class MTOnePlace:Serializable {
    public var place_point:MTOnePoint = MTOnePoint()
    public var place_name:String = ""
    public var place_description:String = ""
    public var place_id:String = ""
    
    public func fromJson(place:JSON?) {
        if place != nil {
            if let lat = place!["point"]["lat"].asDouble {
                self.place_point.lat = lat
            }
            
            if let lng = place!["point"]["lng"].asDouble {
                self.place_point.lng = lng
            }
            
            if let name = place!["name"].asString {
                self.place_name = name
            }
            
            if let description = place!["description"].asString {
                self.place_description = description
            }
            
            if let place_id = place!["place_id"].asString {
                self.place_id = place_id
            }
        }
    }
}


@objc public class MTOneAddress:Serializable {
    var type:String = ""
    var number:String = ""
    var street:String = ""
    var city:String = ""
    var state:String = ""
    var postal:String = ""
    
    public func hasCompleteAddress() -> Bool {
        if self.number != "" && self.street != "" && self.city != "" && self.state != "" && self.postal != "" {
            return true
        } else {
            return false
        }
    }
    
    
    public func fromJson(data: JSON) {
        if let number = data["number"].asString {
            self.number = number
        }
        
        if let street = data["street"].asString {
            self.street = street
        }
        
        if let city = data["city"].asString {
            self.city = city
        }
        
        if let state = data["state"].asString {
            self.state = state
        }
        
        if let postal = data["postal"].asString {
            self.postal = postal
        }
        
        if let type = data["type"].asString {
            self.type = type
        }
    }
}


@objc public class MTOneInfo:Serializable {
    public var expires:String = ""
    public var values:MTOneDictionary = MTOneDictionary()
    public var address:MTOneAddress = MTOneAddress()
    public var place_id:String = ""
    public var beacons:[MTOneBeacon] = []
    public var places:[MTOnePlace] = []
    
    public func fromJson(data: JSON) {
        if let place = data["info"]["place_id"].asString {
            self.place_id = place
        }
        
        if let expires = data["info"]["expires"].asString {
            self.expires = expires
        }
        
        self.places = []
        let place:JSON? = data["info"]["places"][0]
        if place != nil {
            var newPlace:MTOnePlace = MTOnePlace()
            newPlace.fromJson(place)
            self.places.append(newPlace)
        }
        
        
        self.beacons = []
        let beacon:JSON? = data["info"]["beacons"][0]
        if beacon != nil {
            var newBeacon:MTOneBeacon = MTOneBeacon()
            newBeacon.fromJson(beacon)
            
            self.beacons.append(newBeacon)
        }
        
        // Store the address
        self.address.fromJson(data["info"]["address"])
        
        // Pack the values into the values dictionary
        let data = data["info"]["values"]
        self.values.clear()
        
        for (k, outer_value) in data {
            if let name = outer_value["name"].asString {
                if let value = outer_value["value"].asString {
                    self.values.addString(name, value: value)
                }
            }
        }
    }
    
    public func count() -> Int {
        return self.values.count()
    }
}


@objc public class MTOneUser:Serializable {
    var name:String = ""
    var firstName:String = ""
    var lastName:String = ""
    
    var anonymousEmail:String = ""
    
    // If an invalid email address was sent, reject it and keep what was already present.
    var email:String = "" {
        didSet(oldValue) {
            if self.email.rangeOfString("anonymous") != nil {
                if self.email.rangeOfString("mtone-dev.net") != nil {
                    anonymousEmail = email
                    email = "anonymous"
                }
            }
            
            if MTOneSharedInstance.isValidEmail(self.email) == false && email != "anonymous" {
                email = oldValue
            }
        }
    }
    
    var locationIsVisible:Bool = false
    
    var password:String = ""
    var confirmPassword:String = ""
    
    // These are custom fields stored in user_settings.  All user settings must be objects that can be cast to real type.  A Double, Float, Int, Bool (for example) is NSNumber, String is NSString. The client is responsible for storing data in the proper type and querying for the proper type.
    var userSettings:MTOneDictionary = MTOneDictionary()
    
    
    // This reads values from the JSON and initializes the object
    public func fromJSON(data: JSON) {
        self.userSettings = MTOneDictionary()
        
        if let jEmail = data[0]["info"]["email"].asString {
            self.email = jEmail
        }
        
        if let jName = data[0]["info"]["name"].asString {
            self.name = jName
        }
        
        if let jFirstName = data[0]["info"]["first_name"].asString {
            self.firstName = jFirstName
        }
        
        if let jLastName = data[0]["info"]["last_name"].asString {
            self.lastName = jLastName
        }
        
        if let jLocationIsVisible = data[0]["info"]["location_is_visible"].asBool {
            self.locationIsVisible = jLocationIsVisible
        }
        
        if let jUserSettings = data[0]["info"]["user_settings"].asDictionary {
            self.readUserSettings(jUserSettings)
        }
    }
    
    
    private func readUserSettings(userSettings:Dictionary<String, AnyObject>) {
        for (key, value) in userSettings {
            if let j = value as? JSON {
                let n:Double? = j.asDouble
                if n != nil {
                    self.userSettings.addDouble(key, value: n!)
                }
                
                let s:String? = j.asString
                if s != nil {
                    self.userSettings.addString(key, value: s!)
                }
            }
        }
    }
    
    
    // This creates the dictionary that is ued to set values
    override func toJsonString() -> NSString! {
        var newDict:Dictionary<String, AnyObject> = [:]
        
        if email != "" {
            newDict["email"] = email
        }
        
        if firstName != "" {
            newDict["first_name"] = firstName
        }
        
        if lastName != "" {
            newDict["last_name"] = lastName
        }
        
        if firstName != "" {
            newDict["first_name"] = firstName
        }
        
        if password != "" && confirmPassword != "" {
            newDict["password"] = [password, confirmPassword]
        }
        
        self.userSettings.each { (key, value) -> () in
            newDict[key] = value
        }
        
        var dataDict:Dictionary<String, [AnyObject]> = [:]
        var infoDict:Dictionary<String, AnyObject> = [:]
        var infoArray:Dictionary<String, AnyObject> = newDict
        
        infoDict["info"] = infoArray
        dataDict["data"] = [infoDict]
        
        return MTOneSharedInstance.JSONStringify(dataDict)
    }
}


@objc public class MTOneSDK:NSObject, CLLocationManagerDelegate, CBCentralManagerDelegate,UIApplicationDelegate {
    let defaultBaseURL:String = "https://mtone-test.tuner2.net/api/v4/"
    
    // The type of bluetooth beacon filter to be used
    var filter:BeaconFilter = .KnownOnly
    
    // API related
    var baseURL: String
    var apiKey: String = ""
    var token: String?
    
    // Bluetooth related
    var bluetoothAvailable:Bool?
    var cbManager:CBCentralManager?
    var bluetoothRunning = false
    
    // GPS flags
    var GPSAvailable:Bool?
    var GPSRunning = false
    
    // The full UUID for known beacons
    var beacons: [String?] = []
    
    // Proximity UUID that are known
    var proximityBeacons: [String?] = []
    
    // Known beacon regions
    var rangedRegions: Dictionary<String, CLBeaconRegion> = [:]
    
    // List of UUID's to scan
    var proximityUUIDChange: Bool = false
    var proximityUUIDToMonitor:[String] = []
    
    // The nearest beacon that was found
    var nearestBeaconWithActions: CLBeacon? = nil
    var lastBeaconNotification: String = ""
    // This is the beacon being played by the client, set by the client
    var nearestBeacon: CLBeacon? = nil
    
    // Cache for proximity get
    var proximityGetCache: Dictionary<String, ServerResponse> = [:]
    var proximityGetCacheState: Dictionary<String, String> = [:]
    
    var recentProximityGetBeacons:[AnyObject] = []
    
    // The shared location manager
    var locationManager:CLLocationManager = CLLocationManager()
    var locationStatus: String = ""
    var lastLocation: CLLocation? = nil
    
    // The place ID for the last known beacon
    var currentPlace: String? = nil
    var currentLocation: String? = nil
    var currentPlaceInfo: MTOneInfo = MTOneInfo()
    
    // Keeps track of time no beacons were detected
    var timeSinceNoBeacons:NSDate? = nil
    var beaconFailedAtTime:NSDate? = nil
    
    var reachability:Unmanaged<SCNetworkReachability>!
    
    var failedProximityGet:Dictionary<String, Int> = [:]
    
    var clientRunsWhenInUse:Bool = false
    var clientRunsAlways:Bool = false
    
    var clientRunsBluetoothInBackground:Bool = false
    var clientRunsCoreLocationInBackground:Bool = false
    
    var bluetoothErrorAlreadySent:Bool = true
    
    var userInfo:MTOneUser? = nil
    
    // This variable ONLY contains the names of endpoints which are allowed for the user.
    var userPermissions:[String] = []
    
    var userIsAnonymous:Bool = false
    
    var usesBluetoothBeacons:Bool = false
    var usingGPS:Bool = false
    var autoConfigure:Bool = false
    
    class var sharedInstance:MTOneSDK {
        return MTOneSharedInstance
    }
    
    
    override init() {
        NSLog("init")
        
        self.baseURL = self.defaultBaseURL
        
        super.init()
        
        self.initSystem()
        
        // Check the info.plist to make sure the keys are present to allow operation
        
        self.locationManager.delegate = self
        self.cbManager = CBCentralManager(delegate: self, queue: nil)
    }
    
    
    private func initSystem() {
        self.beacons = []
        self.proximityUUIDToMonitor = []
        self.rangedRegions = [:]
        self.nearestBeaconWithActions = nil
        self.proximityGetCache = [:]
        self.proximityGetCacheState = [:]
        self.proximityBeacons = []
        self.token = ""
    }
    
    
    // If both InUse and Always flags are present in the info.plist, the SDK will assume the intent is Always.
    
    private func checkInfoPlist() {
        if let text = NSBundle.mainBundle().infoDictionary?["NSLocationWhenInUseUsageDescription"] as? String {
            if text == "" {
                println("ERROR: The text in NSLocationWhenInUseUsageDescription must not be empty.")
            } else {
                self.clientRunsWhenInUse = true
            }
        }
        
        if let text = NSBundle.mainBundle().infoDictionary?["NSLocationAlwaysUsageDescription"] as? String {
            if text == "" {
                println("ERROR: The text in NSLocationAlwaysUsageDescription must not be empty.")
            } else {
                self.clientRunsAlways = true
            }
        }
        
        
        if let backgroundModes = NSBundle.mainBundle().infoDictionary?["UIBackgroundModes"] as? [String] {
            if contains(backgroundModes, "bluetooth-central") {
                self.clientRunsBluetoothInBackground = true
            }
            
            if contains(backgroundModes, "location") {
                self.clientRunsCoreLocationInBackground = true
            }
        } else {
            println("HINT: The info.plist does not contain entries for background modes.  Your application will only run when it is in use.")
        }
        
        
        if self.clientRunsAlways == true {
            if self.clientRunsBluetoothInBackground == false {
                println("The info.plist seems to indicate the application may run in the background but 'bluetooth-central' is not set in the 'UIBackgroundModes' key.")
                
                // Because corebluetooth is not enabled in the background, this flag is cleared.
                self.clientRunsAlways = false
            }
            
            if self.clientRunsCoreLocationInBackground == false {
                println("The info.plist seems to indicate the application may run in the background but 'location' is not set in the 'UIBackgroundModes' key.")
            }
        }
        
        
        if self.clientRunsWhenInUse == false && self.clientRunsAlways == false && self.usesBluetoothBeacons == true {
            self.showAlertWithOK("ERROR!", message: "IOS8+ REQUIRES either the NSLocationAlwaysUsageDescription or NSLocationWhenInUseUsageDescription set in the projects info.plist.  It must contain appropriate string for the user to view when the app asks permission to enable Bluetooth.  Additionally if you have chosen to set NSLocationAlwaysUsageDescription you MUST include the UIBackgroundModes and add bluetooth-central and location")
        }
    }
    
    
    public func isConnectedToNetwork() -> Bool {
        var zeroAddress = sockaddr_in(sin_len: 0, sin_family: 0, sin_port: 0, sin_addr: in_addr(s_addr: 0), sin_zero: (0, 0, 0, 0, 0, 0, 0, 0))
        zeroAddress.sin_len = UInt8(sizeofValue(zeroAddress))
        zeroAddress.sin_family = sa_family_t(AF_INET)
        
        let defaultRouteReachability = withUnsafePointer(&zeroAddress) {
            SCNetworkReachabilityCreateWithAddress(nil, UnsafePointer($0)).takeRetainedValue()
        }
        
        var flags: SCNetworkReachabilityFlags = 0
        if SCNetworkReachabilityGetFlags(defaultRouteReachability, &flags) == 0 {
            return false
        }
        
        let isReachable = (flags & UInt32(kSCNetworkFlagsReachable)) != 0
        let needsConnection = (flags & UInt32(kSCNetworkFlagsConnectionRequired)) != 0
        
        return (isReachable && !needsConnection) ? true : false
    }
    
    
    public func callMTOneAsPost(endpoint: String, params: Dictionary<String, String>, completion: (response: ServerResponse) -> Void) {
        var backgroundQueue = NSOperationQueue()
        backgroundQueue.addOperationWithBlock() {
            var request = NSMutableURLRequest(URL: NSURL(string: self.baseURL + endpoint + "/")!)
            var session = NSURLSession.sharedSession()
            request.HTTPMethod = "POST"
            
            var bodyData = self.dictionaryToParameterList(params)
            request.HTTPBody = bodyData.dataUsingEncoding(NSUTF8StringEncoding, allowLossyConversion: false)
            
            request.addValue("application/x-www-form-urlencoded", forHTTPHeaderField: "Content-Type")
            request.addValue("application/x-www-form-urlencoded", forHTTPHeaderField: "Accept")
            
            var task = session.dataTaskWithRequest(request, completionHandler: {data, response, error -> Void in
                var strData = NSString(data: data, encoding: NSUTF8StringEncoding)
                
                var error: NSError?
                var json = NSJSONSerialization.JSONObjectWithData(data, options: .MutableLeaves, error: &error) as? NSDictionary
                
                // Did the JSONObjectWithData constructor return an error? If so, log the error to the console
                if(error != nil) {
                    println(error!.localizedDescription)
                    
                    let theResponse = ServerResponse(success: false, json: JSON(""), status: 0, error: error);
                    dispatch_async(dispatch_get_main_queue(), {
                        completion(response: theResponse)
                    })
                    return
                } else {
                    var json = JSON.parse(strData!)
                    
                    let status: Int? = json["response"]["information"]["code"].asInt
                    
                    // The status field is missing, return a failure
                    if status == nil {
                        let theResponse = ServerResponse(success: false, json: json, status: 0, error: error);
                        dispatch_async(dispatch_get_main_queue(), {
                            completion(response: theResponse)
                        })
                        return
                    }
                    
                    
                    // The status is more than 300, usually a bad sign
                    if status >= 300 {
                        let theResponse = ServerResponse(success: false, json: json, status: status!, error: error);
                        dispatch_async(dispatch_get_main_queue(), {
                            completion(response: theResponse)
                        })
                        return
                    }
                    
                    
                    let theResponse = ServerResponse(success: true, json: json, status: status!, error: error);
                    
                    // Populate the response with something useful
                    theResponse.action = json["response"]["action"]
                    theResponse.data = json["response"]["data"]
                    
                    dispatch_async(dispatch_get_main_queue(), {
                        completion(response: theResponse)
                    })
                    return
                }
            })
            
            task.resume()
        }
    }
    
    
    public func callMTOne(endpoint: String, params: Dictionary<String, String>, completion: (response: ServerResponse) -> Void) {
        var backgroundQueue = NSOperationQueue()
        backgroundQueue.addOperationWithBlock() {
            var newParams:Dictionary<String, String> = params
            newParams["timestamp"] =  "\(NSDate().timeIntervalSince1970)"
            
            let url = self.baseURL + endpoint + "/?" + self.dictionaryToParameterList(newParams)
            let json = JSON.fromURL(url)
            var error:NSError? = nil
            
            if let error = json.asError {
            }
            
            
            // The call failed badly, return some information about it.
            if error != nil {
                println(error!.localizedDescription)
                
                let theResponse = ServerResponse(success: false, json: json, status: 0, error: error);
                dispatch_async(dispatch_get_main_queue(), {
                    completion(response: theResponse)
                })
                return
            }
            
            let status: Int? = json["response"]["information"]["code"].asInt
            
            // The status field is missing, return a failure
            if status == nil {
                let theResponse = ServerResponse(success: false, json: json, status: 0, error: error);
                dispatch_async(dispatch_get_main_queue(), {
                    completion(response: theResponse)
                })
                return
            }
            
            
            // The status is more than 300, usually a bad sign
            if status >= 300 {
                let theResponse = ServerResponse(success: false, json: json, status: status!, error: error);
                dispatch_async(dispatch_get_main_queue(), {
                    completion(response: theResponse)
                })
                return
            }
            
            
            let theResponse = ServerResponse(success: true, json: json, status: status!, error: error);
            
            // Populate the response with something useful
            theResponse.action = json["response"]["action"]
            theResponse.data = json["response"]["data"]
            
            dispatch_async(dispatch_get_main_queue(), {
                completion(response: theResponse)
            })
            return
        }
    }
    
    
    private func createFailedResponse() -> ServerResponse {
        var response:ServerResponse = ServerResponse(success: false, json: JSON(""), status: 401, error: nil)
        
        return response
    }
    
    
    public func authCreateUser(email: String, password: String, confirm: String,completion: (response: ServerResponse)-> Void) -> Void {
        if email != "" && password != "" && confirm != "" {
            let parms = ["username": email, "password": password, "password_confirm" : confirm, "key": self.apiKey] as Dictionary<String, String>
            
            self.callMTOne("auth/create_user", params: parms, completion: {
                (response:ServerResponse) in
                
                self.token = ""
                
                if response.success == false {
                    response.showError()
                    
                    let false_response = self.createFailedResponse()
                    completion(response: false_response)
                    
                    return
                }
                
                completion(response: response)
            })
        } else {
            var missingFields:[String] = []
            if email == "" {
                missingFields.append("email")
            }
            
            if password == "" {
                missingFields.append("password")
            }
            
            if confirm == "" {
                missingFields.append("confirm")
            }
            
            var message = join(", ", missingFields) + " are required to sign up."
            
            self.showAlertWithOK("Error", message: message)
            
            let response = self.createFailedResponse()
            completion(response: response)
        }
    }
    
    
    public func authResetPassword(email: String, completion: (response: ServerResponse)-> Void) -> Void {
        if email != "" {
            let parms = ["username": email, "key": self.apiKey] as Dictionary<String, String>
            
            self.callMTOne("auth/reset_password", params: parms, completion: {
                (response:ServerResponse) in
                
                self.token = ""
                
                if response.success == false {
                    response.showError()
                    
                    let false_response = self.createFailedResponse()
                    completion(response: false_response)
                    
                    return
                }
                
                completion(response: response)
            })
        } else {
            self.showAlertWithOK("Error", message: "You must enter an email address.")
            
            let response = self.createFailedResponse()
            completion(response: response)
        }
    }
    
    
    public func appShutDown() {
        // Turn off the services.  They may have been turned on during app execution so they should now be turned off.
        self.stopBlueTooth()
        self.stopGPS()
        
        // Send any log entries that may be left
        MTOneSharedLogger.sendEntriesToServer(force: true)
        MTOneSharedLogger.clear()
    }
    
    
    public func authLogout(completion: (response: ServerResponse)-> Void) -> Void {
        if let token:String = self.token {
            let parms = ["token": token] as Dictionary<String, String>
            
            self.appShutDown()
            
            self.callMTOne("auth/logout", params: parms, completion: {
                (response:ServerResponse) in
                
                // This clears the values so the object is reset
                self.initSystem()
                
                self.saveToken()
                
                completion(response: response)
            })
        }
    }
    
    
    // Unlike authLogout which turns off GPS and Bluetooth, this does not turn them on because an application may not need them.
    public func authLogin(email: String, password: String, completion: (response: ServerResponse)-> Void) {
        if email == "anonymous" {
            self.userIsAnonymous = true
            
            if password != "" {
                return self.errorAlert("Anonymous users must send an empty password")
            }
            
            self.isValidToken({ (response) -> Void in
                if response == true {
                    var response = ServerResponse(success: true, json: JSON(""), status: 200, error: nil)
                    return completion(response: response)
                } else {
                    self.internalAuthLogin(email, password: password, completion: completion)
                }
            })
            
            return
        } else {
            self.userIsAnonymous = false
        }
        
        if email == "" && password == "" {
            return self.errorAlert("You must provide email and password")
        }
        
        if email == "" {
            return self.errorAlert("Email is required")
        }
        
        if password == "" {
            return self.errorAlert("Password is required")
        }
        
        self.internalAuthLogin(email, password: password, completion: completion)
    }
    
    
    private func internalAuthLogin(email: String, password: String, completion: (response: ServerResponse) -> Void) {
        let parms = ["username": email, "password": password, "key": self.apiKey] as Dictionary<String, String>
        
        self.callMTOne("auth/login", params: parms, completion: {
            (response:ServerResponse) in
            
            self.userPermissions = []
            
            self.token = nil
            if response.success {
                // Get the token
                if let token:String? = response.json!["response"]["data"][0]["token"].asString {
                    self.token = token
                    
                    // Now store the permissions array
                    var permissions:JSON? = response.json!["response"]["data"][0]["permissions"]
                    if permissions != nil {
                        for (k, inner_data) in permissions! {
                            if let inner_id:String = inner_data.asString {
                                self.userPermissions.append(inner_id)
                            }
                        }
                    }
                    
                    
                    self.saveToken()
                } else {
                    response.success = false
                }
            }
            
            completion(response: response)
        })
    }
    
    
    public func userHasPermission(permissionName: String) -> Bool {
        if contains(self.userPermissions, permissionName) {
            return true
        } else {
            return false
        }
    }
    
    
    public func userFavoriteSet(favorite: MTOneFavoriteAction, tag:String, completion: (response: ServerResponse)-> Void) {
        if let token:String = self.token {
            let favoriteJSON:String = favorite.toJsonString()
            
            let parms = ["service": "default", "token": token, "favorite": favoriteJSON, "tag": tag] as Dictionary<String, String>
            
            self.callMTOne("user/favorite_set", params: parms, completion: {
                (response:ServerResponse) in
                
                completion(response: response)
            })
        }
    }
    
    
    public func userFavoriteList(type: String, completion: (response: ServerResponse)-> Void) {
        if let token:String = self.token {
            let parms = ["service": "default", "token": token, "type": type] as Dictionary<String, String>
            
            self.callMTOne("user/favorite_list", params: parms, completion: {
                (response:ServerResponse) in
                
                completion(response: response)
            })
        }
    }
    
    
    public func userFavoriteDelete(globalItemId: String, completion: (response: ServerResponse)-> Void) {
        if let token:String = self.token {
            let parms = ["service": "default", "token": token, "global_item_id": globalItemId] as Dictionary<String, String>
            
            self.callMTOne("user/favorite_delete", params: parms, completion: {
                (response:ServerResponse) in
                
                completion(response: response)
            })
        }
    }
    
    
    
    public func authGNDN(completion: (response: ServerResponse)-> Void) {
        if let token:String = self.token {
            let parms = ["service": "default", "token": token] as Dictionary<String, String>
            
            self.callMTOne("auth/gndn", params: parms, completion: {
                (response:ServerResponse) in
                
                completion(response: response)
            })
        }
    }
    
    
    public func userInfoGet(completion: (response: ServerResponse)-> Void) {
        if let token:String = self.token {
            let parms = ["service": "default", "token": token] as Dictionary<String, String>
            
            self.callMTOne("user/info_get", params: parms, completion: {
                (response:ServerResponse) in
                
                if self.userInfo == nil {
                    self.userInfo = MTOneUser()
                }
                
                if response.success == true {
                    self.userInfo!.fromJSON(response.data!)
                    
                    completion(response: response)
                } else {
                    let false_response = self.createFailedResponse()
                    completion(response: false_response)
                }
            })
        }
    }
    
    
    // May need to validate data before this is called
    public func userInfoSet(completion: (response: ServerResponse)-> Void) {
        if userInfo == nil {
            self.showAlertWithOK("Error", message: "userInfoSet was called before userInfoGet.  There is nothing to do.")
        }
        
        if let token:String = self.token {
            let user_as_param_string = userInfo!.toJsonString()
            
            let parms = ["service": "default", "token": token, "parameters": user_as_param_string] as Dictionary<String, String>
            
            self.callMTOneAsPost("user/info_set", params: parms, completion: {
                (response:ServerResponse) in
                
                if response.success == true {
                    completion(response: response)
                } else {
                    response.showError()
                    
                    let false_response = self.createFailedResponse()
                    completion(response: false_response)
                }
            })
        }
    }
    
    
    // The attachment dictioary should use mime_type as the key and NSData as the dictionary
    public func proximitySet(beacon:MTOneProximityBeacon?, location:MTOneProximityLocation?, place:MTOneProximityPlace?, completion: (response: ServerResponse)-> Void) {
        
        // Exit immediately if the user cannot call this endpoint.
        if self.userHasPermission("proximity_set") == false {
            let false_response = self.createFailedResponse()
            return completion(response: false_response)
        }
        
        if let token:String = self.token {
            var parms = ["service": "mtoneproximity", "token": token] as Dictionary<String, String>
            
            if beacon != nil {
                let beaconJSON = self.JSONStringify(beacon!.toDictionary())
                parms["beacon"] = beaconJSON
            }
            
            if location != nil {
                let locationJSON = self.JSONStringify(location!.toDictionary())
                parms["location"] = locationJSON
            }
            
            if place != nil {
                let placeJSON = self.JSONStringify(place!.toDictionary())
                parms["place"] = placeJSON
            }
            
            // Make a POST version of this so the server can handle the data sizes sent.
            self.callMTOneAsPost("proximity/set", params: parms, completion: {
                (response:ServerResponse) in
                
                if response.success == true {
                    self.recentProximityGetBeacons = []
                    
                    // Now clear the cache and refresh so the data is current
                    var uuids = Array(self.proximityGetCacheState.keys)
                    
                    self.proximityGetCache = [:]
                    self.proximityGetCacheState = [:]
                    
                    self.proximityGet(uuids, completion: { (response) -> Void in
                        completion(response: response)
                    })
                } else {
                    completion(response: response)
                }
            })
        }
    }
    
    
    public func proximityList(completion: (response: ServerResponse)-> Void) {
        if let token:String = self.token {
            let parms = ["service": "mtoneproximity", "token": token] as Dictionary<String, String>
            
            self.callMTOne("proximity/list", params: parms, completion: {
                (response:ServerResponse) in
                
                self.proximityBeacons = []
                
                if let data = response.data {
                    for (k, inner_data) in data {
                        if let inner_id = inner_data["value"].asString {
                            self.proximityBeacons.append(inner_id)
                        }
                    }
                }
                
                completion(response: response)
            })
        }
    }
    
    
    public func proximityGet(id:[String], completion: (response: ServerResponse)-> Void) {
        if id.isEmpty == true {
            return;
        }
        
        dispatch_async(dispatch_get_main_queue(), {
            let first_id:String = id[0]
            
            if let state:String? = self.proximityGetCacheState[first_id] {
                if state == "fetching" {
                    return
                }
                
                if state == "found" {
                    if let response:ServerResponse? = self.proximityGetCache[first_id] {
                        completion(response: response!)
                        return
                    }
                }
            }
            
            if let token:String = self.token {
                var accepted_ids = id
                
                // accepted_ids = self.removeFailedProximityGet(id)
                
                self.setProximityState(accepted_ids, state: "fetching")
                
                var zone_list:[String] = []
                if self.recentProximityGetBeacons.count > 0 {
                    for a_id in 0..<accepted_ids.count {
                        let a_full_id = accepted_ids[a_id]
                        
                        for b_id in 0..<self.recentProximityGetBeacons.count {
                            let b_beacon = self.recentProximityGetBeacons[b_id] as? CLBeacon
                            
                            if b_beacon != nil {
                                let b_full_id = self.makeFullUUID(b_beacon)
                                
                                if a_full_id == b_full_id {
                                    let zone = self.getBeaconProximityAsString(b_beacon!)
                                    zone_list.append(zone)
                                    break
                                }
                            }
                        }
                    }
                }
                
                //JEM id needs to be converted to a string
                let ids:String = self.JSONStringify(accepted_ids)
                let zones:String = self.JSONStringify(zone_list)
                
                var lat:Double = 0.0
                var lng:Double = 0.0
                
                if self.lastLocation != nil {
                    lat = self.lastLocation!.coordinate.latitude
                    lng = self.lastLocation!.coordinate.longitude
                }
                
                let parms = ["id": ids, "service": "mtoneproximity", "token": token, "lat" :self.doubleToString(lat), "lng": self.doubleToString(lng), "zone": zones, "auto_configure": self.boolToString(self.autoConfigure)] as Dictionary<String, String>
                
                self.callMTOne("proximity/get", params: parms, completion: {
                    (response:ServerResponse) in
                    
                    if response.success == true {
                        // Set the list to 404, the next block will then set the status back to the appropriate one if the data was found
                        self.setProximityState(accepted_ids, state: "404")
                        
                        // The response will be an array, we need to break that up into individual pieces and store each as a successful operation
                        if let actions = response.action {
                            for (k, inner_action) in actions {
                                if let inner_ids = inner_action["info"]["beacons"].asArray {
                                    for inner_id in inner_ids {
                                        if let the_inner_id:String? = inner_id["beacon_id"].asString {
                                            if the_inner_id != nil {
                                                if self.proximityGetCacheHas(the_inner_id!) == false {
                                                    var new_response = ServerResponse(success: response.success, json: response.json!, status: response.status!, error: response.error)
                                                    
                                                    new_response.action = response.action
                                                    new_response.data = response.data
                                                    new_response.action = inner_action
                                                    
                                                    var id_as_array:[String] = []
                                                    id_as_array.append(the_inner_id!)
                                                    self.setProximityState(id_as_array, state: "found")
                                                    
                                                    self.proximityGetCache[the_inner_id!] = new_response
                                                    
                                                    // Keep track of the full UUID
                                                    self.beacons.append(the_inner_id)
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            
                            self.checkForFailedProximityGet(accepted_ids)
                            
                            completion(response: response)
                        }
                    } else {
                        NSLog("Beacon not found %@", accepted_ids)
                    }
                })
            }
        })
    }
    
    
    // Remove failed ID's so they don't cause calls to the server.//
    private func removeFailedProximityGet(ids: [String]) -> [String] {
        var keptIDS:[String] = []
        
        for i in 0..<ids.count {
            var id:String = ids[i]
            var state = self.getProximityState(id)
            
            if state == "failed" && self.failedProximityGet[id] > 5 {
                var count = self.failedProximityGet[id]
                println("\(id) NOT FOUND \(count)")
            } else {
                keptIDS.append(id)
            }
        }
        
        
        return keptIDS
    }
    
    
    private func checkForFailedProximityGet(ids: [String]) {
        // If the time since failure is more than a day, clear the whole cache out
        if self.beaconFailedAtTime != nil {
            let elapsedTime = NSDate().timeIntervalSinceDate(self.beaconFailedAtTime!)
            
            if elapsedTime > 86240 {
                self.failedProximityGet = [:]
                self.beaconFailedAtTime = nil
            }
        }
        
        for uuid in ids {
            let state = self.proximityGetCacheState[uuid]
            
            if state == "404" {
                let index = failedProximityGet.indexForKey(uuid)
                
                if index == nil {
                    failedProximityGet[uuid] = 1
                } else {
                    failedProximityGet[uuid] = failedProximityGet[uuid]! + 1
                    
                    if failedProximityGet[uuid] > 5 {
                        var arrayed_id:[String] = []
                        arrayed_id.append(uuid)
                        self.setProximityState(arrayed_id, state: "failed")
                        
                        if self.timeSinceNoBeacons == nil {
                            self.timeSinceNoBeacons = NSDate()
                        }
                    }
                }
            }
        }
    }
    
    
    private func proximityGetCacheHas(uuid: String) -> Bool {
        if self.proximityGetCache.indexForKey(uuid) != nil {
            return true
        }
        
        return false
    }
    
    
    private func getProximityState(id: String) -> String {
        let index = self.proximityGetCacheState.indexForKey(id)
        if index == nil {
            return "unknown"
        } else {
            return self.proximityGetCacheState[id]!
        }
    }
    
    
    private func setProximityState(ids: [String], state: String) {
        for i in 0..<ids.count {
            let uuid = ids[i]
            
            let old_state = self.getProximityState(uuid)
            
            if old_state != "found" {
                println("setProximityState \(uuid) = \(state)")
                
                let index = self.proximityGetCacheState.indexForKey(uuid)
                if index != nil {
                    self.proximityGetCacheState.updateValue(state, forKey: uuid)
                } else {
                    self.proximityGetCacheState[uuid] = state
                }
            }
        }
    }
    
    
    public func getActionForBeaconId(beaconId: String) -> JSON? {
        return self.proximityGetCache[beaconId]?.action
    }
    
    
    public func getLocationID(action: JSON?) -> String {
        var result = ""
        
        if action != nil {
            if let location = action?["info"]["location_id"].asString {
                result = location
            }
        }
        
        return result
    }
    
    
    public func getInfoValueForKey(key: String) -> String {
        return self.currentPlaceInfo.values.asString(key)
    }
    
    
    public func callProximityListAndStartBluetooth() {
        self.usesBluetoothBeacons = true
        self.checkInfoPlist()
        
        self.proximityList({
            (response) in
            
            var proximityUUID:[String] = []
            
            if let data = response.data {
                for (k, v) in data {
                    if let uuid = v["value"].asString {
                        proximityUUID.append(uuid)
                    }
                }
                
                self.proximityUUIDToMonitor = proximityUUID
                self.startBlueTooth()
                self.startGPS()
            }
        })
    }
    
    
    public func getActionsForThisUUIDAndZone(fullUUID: String, zone: String) -> Dictionary<String, JSON> {
        var result:Dictionary<String, JSON> = [:]
        
        var arrayed_FullUUID:[String] = []
        
        arrayed_FullUUID.append(fullUUID)
        
        if let action = self.proximityGetCache[fullUUID]?.action {
            var id:Int = 0
            
            let actions = action["actions"]
            
            for (k, v) in actions {
                if let handler = v["handler"].asString {
                    if handler == zone {
                        let keep:JSON = v
                        result[String(++id)] = keep
                    }
                }
            }
        }
        
        return result
    }
    
    
    public func getActionsForThisBeacon(beacon: CLBeacon) -> Dictionary <String, JSON> {
        var result:Dictionary<String, JSON> = [:]
        
        var arrayed_FullUUID:[String] = []
        var fullUUID = self.makeFullUUID(beacon)
        
        arrayed_FullUUID.append(fullUUID)
        
        if let action = self.proximityGetCache[fullUUID]?.action {
            var id:Int = 0
            
            let actions = action["actions"]
            
            for (k, v) in actions {
                if let handler = v["handler"].asString {
                    if handler == self.getBeaconProximityAsString(beacon) {
                        let keep:JSON = v
                        result[String(++id)] = keep
                    }
                }
            }
        }
        
        return result
    }
    
    
    public func getActionsForNearestBeacon() -> Dictionary<String, JSON> {
        if let currentBeacon:CLBeacon = MTOneSharedInstance.nearestBeaconWithActions {
            return self.getActionsForThisBeacon(currentBeacon)
        }
        
        var result:Dictionary<String, JSON> = [:]
        return result
    }
    
    
    public func getBeaconProximityAsString(beacon: CLBeacon) -> String {
        var result:String = "unknown"
        
        switch (beacon.proximity) {
        case .Near:
            result = "near"
            break
        case .Far:
            result = "far"
            break
        case .Immediate:
            result = "immediate"
            break
        case .Unknown:
            result = "unknown"
            break
        }
        
        return result
    }
    
    
    public func errorAlert(message: String) {
        self.showAlertWithOK("Error", message: message)
    }
    
    
    public func showAlertWithOK(title: String, message: String) {
        var alert: UIAlertView
        
        alert = UIAlertView();
        alert.title = title
        alert.message = message
        alert.addButtonWithTitle("OK")
        alert.show()
    }
    
    
    //BLUETOOTH
    public func isBlueToothAvailable() -> Bool {
        if (self.cbManager != nil) {
            // Prompt with the appropriate access level (ALWAYS)
            if self.clientRunsAlways {
                self.locationManager.requestAlwaysAuthorization()
            } else if self.clientRunsWhenInUse {
                self.locationManager.requestWhenInUseAuthorization()
            } else {
                println("CHECK THE LOG! There is a configuration issue in the applications info.plist")
            }
            
            let state:CBCentralManagerState? = self.cbManager?.state
            
            if state == CBCentralManagerState.Unknown {
                self.bluetoothAvailable = true
            } else if(state == CBCentralManagerState.Unauthorized) {
                self.bluetoothAvailable = false
                self.errorAlert("Bluetooth access is not authorized")
            } else if(state == CBCentralManagerState.Unsupported) {
                self.bluetoothAvailable = false
                if self.bluetoothErrorAlreadySent == false {
                    self.errorAlert("Your device doesn't support bluetooth")
                    self.bluetoothErrorAlreadySent = true
                }
            } else if(state == CBCentralManagerState.Resetting) {
                self.bluetoothAvailable = false
            } else if(state == CBCentralManagerState.PoweredOff) {
                self.bluetoothAvailable = false
                self.errorAlert("Bluetooth must be turned on")
            } else if(state == CBCentralManagerState.PoweredOn) {
                self.bluetoothAvailable = true
            }
        }
        
        return self.bluetoothAvailable!
    }
    
    
    public func startBlueTooth() {
        if self.isBlueToothAvailable() == true {
            self.beacons = []
            
            // Populate the regions we will range once.
            self.rangedRegions = [:]
            
            for uuid:String in self.proximityUUIDToMonitor {
                if let proximityUUID:NSUUID? = NSUUID(UUIDString: uuid) {
                    var aRegion:CLBeaconRegion = CLBeaconRegion(proximityUUID:proximityUUID, identifier:proximityUUID!.UUIDString);
                    aRegion.notifyEntryStateOnDisplay = true
                    aRegion.notifyOnEntry = true
                    aRegion.notifyOnExit = false
                    self.rangedRegions[uuid] = aRegion
                    self.locationManager.startMonitoringForRegion(aRegion)
                    self.locationManager.startRangingBeaconsInRegion(aRegion)
                }
            }
        }
    }
    
    
    public func stopBlueTooth() {
        self.isBlueToothAvailable();
        
        if (self.bluetoothAvailable == true) {
            // Stop ranging when the view goes away.
            for region in self.rangedRegions {
                //                if let aRegion:CLBeaconRegion = region as? CLBeaconRegion {
                //                    self.locationManager!.stopRangingBeaconsInRegion(region: aRegion);
                //                }
            }
            
        }
    }
    
    
    public func centralManagerDidUpdateState(central: CBCentralManager!)
    {
        if self.cbManager != nil {
            switch(self.cbManager!.state) {
            case CBCentralManagerState.PoweredOn:
                self.bluetoothAvailable = true
                
            case .PoweredOff:
                self.bluetoothAvailable = false
                
            case .Unauthorized:
                self.bluetoothAvailable = false
                
            case .Unsupported:
                self.bluetoothAvailable = false
                
            case .Unknown:
                self.bluetoothAvailable = true
                
            case .Resetting:
                self.bluetoothAvailable = false
                
            default:
                self.bluetoothAvailable = false
            }
        }
    }
    
    
    //GPS
    public func isGPSAvailable() -> Bool {
        self.GPSAvailable = true
        
        return self.GPSAvailable!
    }
    
    
    public func startGPS() {
        self.isGPSAvailable();
        
        if (self.GPSAvailable == true) {
            self.usingGPS = true
            
            locationManager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters
            locationManager.distanceFilter = kCLDistanceFilterNone
            locationManager.pausesLocationUpdatesAutomatically = false
            locationManager.startUpdatingLocation()
        }
    }
    
    
    public func stopGPS() {
        self.isGPSAvailable();
        
        if (self.GPSAvailable == true) {
            self.usingGPS = false
            locationManager.stopUpdatingLocation()
        }
    }
    
    
    public func locationManager(manager: CLLocationManager!, didFailWithError error: NSError!) {
        NSLog("Error" + error.description)
    }
    
    
    public func locationManager(manager:CLLocationManager, didUpdateLocations locations:[AnyObject]) {
        if locations.isEmpty == false {
            lastLocation = locations[0] as? CLLocation
        }
    }
    
    
    public func locationManager(manager: CLLocationManager!, didChangeAuthorizationStatus status: CLAuthorizationStatus) {
        switch status {
        case CLAuthorizationStatus.Restricted:
            locationStatus = "Access: Restricted"
            break
        case CLAuthorizationStatus.Denied:
            locationStatus = "Access: Denied"
            NSNotificationCenter.defaultCenter().postNotificationName("CoreLocationDenied", object: nil)
            break
        case CLAuthorizationStatus.NotDetermined:
            locationStatus = "Access: NotDetermined"
            break
        default:
            locationStatus = "Access: Allowed"
        }
    }
    
    
    // JSON and parameter functions
    public func JSONStringify(jsonObj: AnyObject) -> String {
        var e: NSError?
        let jsonData = NSJSONSerialization.dataWithJSONObject(jsonObj, options: NSJSONWritingOptions(0), error: &e)
        if e != nil {
            return ""
        } else {
            return NSString(data: jsonData!, encoding: NSUTF8StringEncoding)!
        }
    }
    
    
    // Convert a dictionary to a URI style parameter list with encoding
    private func dictionaryToParameterList(params: Dictionary<String, String>) -> String {
        var result: String = ""
        
        for (key, value) in params {
            var escapedValue = value.stringByAddingPercentEscapesUsingEncoding(NSUTF8StringEncoding)
            result += "\(key)=" + escapedValue! + "&"
        }
        
        return result
    }
    
    
    public func isRunningInBackground() -> Bool {
        if UIApplication.sharedApplication().applicationState == .Background {
            return true
        } else {
            return false
        }
    }
    
    
    public func isBackgroundThread() -> Bool {
        if (NSRunLoop.currentRunLoop() != NSRunLoop.mainRunLoop()) {
            return true
        } else {
            return false
        }
    }
    
    
    func noBeacon() {
        // exit on the first timestamp
        if self.timeSinceNoBeacons == nil {
            self.timeSinceNoBeacons = NSDate()
            
            return
        }
        
        let currentDate = NSDate()
        
        let difference = currentDate.timeIntervalSince1970 - self.timeSinceNoBeacons!.timeIntervalSince1970
        
        if (difference > 5) {
            self.nearestBeaconWithActions = nil
            
            MTOneSharedLogger.noBeacon()
            
            NSNotificationCenter.defaultCenter().postNotificationName("noBeacons", object:nil)
        }
        
    }
    
    
    private func filterKnownOnlyMatchesProximity(beacons: [CLBeacon?]) -> [CLBeacon?] {
        // Start by removing unknown beacons
        var tempBeacons = self.filterKnownOnly(beacons)
        
        // Then see which beacons match the proximity
        // TODO
        return tempBeacons
    }
    
    
    private func filterKnownOnly(beacons: [CLBeacon?]) -> [CLBeacon?] {
        if beacons.isEmpty == true {
            return beacons
        }
        
        
        var knownBeacons:[CLBeacon?] = []
        
        // This won't work if you move from one set of beacons to another because now proximity get has been called for the new set of beacons
        
        // Now scan the beacons to see which have known full UUID strings
        for foundBeacon in beacons {
            let fFullUUID = self.makeFullUUID(foundBeacon!)
            
            if let state:String? = self.proximityGetCacheState[fFullUUID] {
                if state == "found" {
                    knownBeacons.append(foundBeacon)
                }
            }
        }
        
        return knownBeacons
    }
    
    
    // Look inside for action
    private func beaconHasActions(aBeacon: CLBeacon) -> Bool {
        let actions = self.getActionsForThisBeacon(aBeacon)
        
        if actions.count > 0 {
            return true
        } else {
            return false
        }
    }
    
    
    public func locationManager(manager: CLLocationManager!, didRangeBeacons beacons: [AnyObject]!, inRegion region: CLBeaconRegion!) {
        dispatch_async(dispatch_get_main_queue(), {
            
            // If there are no beacons in the list, exit now
            if beacons.isEmpty == true {
                self.noBeacon()
                return
            }
            
            
            // Remove beacons with RSSI of 0
            var initialTempBeacons:[CLBeacon?] = []
            var ids:[String] = []
            for beacon in beacons {
                let aBeacon:CLBeacon = beacon as CLBeacon
                if beacon.rssi < 0 {
                    initialTempBeacons.append(aBeacon)
                    ids.append(self.makeFullUUID(aBeacon))
                }
            }
            
            ids = self.removeFailedProximityGet(ids)
            
            
            // Start the wrapper here
            // Remember the list of beacons so we can extract the zones later
            self.recentProximityGetBeacons = beacons
            
            MTOneSharedInstance.proximityGet(ids, completion: {
                (response) in
                
                var tempBeacons:[CLBeacon?] = []
                
                switch self.filter {
                case .KnownOnly:
                    tempBeacons = self.filterKnownOnly(initialTempBeacons)
                default:
                    tempBeacons = initialTempBeacons
                }
                
                
                if tempBeacons.isEmpty == true {
                    self.noBeacon()
                    return
                }
                
                self.timeSinceNoBeacons = nil
                
                // The nearest beacon is initially the first beacon
                if var tNearestBeacon:CLBeacon? = tempBeacons[0] as CLBeacon? {
                    // Then we scan and pick a better one if it exists
                    for beacon in tempBeacons {
                        if let currentBeacon:CLBeacon? = beacon as CLBeacon? {
                            if currentBeacon?.rssi > tNearestBeacon?.rssi {
                                tNearestBeacon = currentBeacon;
                            }
                        }
                    }
                    
                    self.nearestBeacon = tNearestBeacon!
                    
                    let id = self.makeFullUUID(tNearestBeacon!)
                    
                    if self.nearestBeaconWithActions != nil {
                        // Don't respond if the same beacon and strenth was detected
                        if id == self.makeFullUUID(self.nearestBeaconWithActions!) {
                            if self.getBeaconProximityAsString(self.nearestBeaconWithActions!) == self.getBeaconProximityAsString(tNearestBeacon!) {
                                
                                if self.nearestBeaconWithActions?.rssi != tNearestBeacon?.rssi {
                                    self.nearestBeaconWithActions = tNearestBeacon
                                    NSNotificationCenter.defaultCenter().postNotificationName("beaconUpdated", object:nil)
                                    
                                    return
                                } else {
                                    return
                                }
                            }
                        }
                    }
                    
                    
                    let lastNearestBeacon = self.nearestBeaconWithActions
                    
                    var newNearestBeacon = false
                    if tNearestBeacon != nil {
                        if self.beaconHasActions(tNearestBeacon!) {
                            self.nearestBeaconWithActions = tNearestBeacon
                            newNearestBeacon = true
                        }
                    }
                    
                    
                    if lastNearestBeacon == nil || self.makeFullUUID(lastNearestBeacon!) != self.makeFullUUID(tNearestBeacon!) {
                        NSNotificationCenter.defaultCenter().postNotificationName("newBeacon", object:nil)
                    }
                    
                    NSNotificationCenter.defaultCenter().postNotificationName("beaconUpdated", object:nil)
                    
                    var id_array:[String] = []
                    id_array.append(id)
                    
                    MTOneSharedInstance.proximityGet(id_array, completion: {
                        (response) in
                        
                        if let action = response.action {
                            self.currentPlaceInfo.fromJson(action)
                            
                            if self.currentPlaceInfo.count() > 0 {
                                NSNotificationCenter.defaultCenter().postNotificationName("placeInfoUpdated", object:nil)
                            }
                            
                            if let place = action["info"]["place_id"].asString {
                                self.currentPlace = place
                            }
                            
                            if let location = action["info"]["location_id"].asString {
                                self.currentLocation = location
                            }
                        }
                        
                        
                        if tNearestBeacon != nil {
                            let beacon_id = self.makeFullUUID(tNearestBeacon)
                            let beacon_zone = self.getBeaconProximityAsString(tNearestBeacon!)
                            let actions = self.getActionsForThisBeacon(tNearestBeacon!)
                            let handler_count = actions.count
                            
                            MTOneSharedLogger.addBeaconEntry(self.lastLocation?, beacon_id:beacon_id, beacon_zone: beacon_zone, location_id: self.currentLocation!, place_id: self.currentPlace!, handler_count: handler_count, other_beacons:tempBeacons)
                            
                        }
                        
                        
                        if newNearestBeacon == true {
                            NSNotificationCenter.defaultCenter().postNotificationName("proximtyGetSuccess", object:nil)
                        }
                    })
                }
            })
        })
    }
    
    
    // Returns an array beacons for a give place ID.  self.currentPlace can be used to find the currently associated beacons
    public func getBeaconsForPlace(aPlaceID: String) -> [JSON] {
        var result:[JSON] = []
        var alreadyReturned:Dictionary<String, String> = [:]
        
        if self.proximityGetCache.isEmpty {
            return result
        }
        
        for (k,v) in self.proximityGetCacheState {
            if v == "found" {
                if let response:ServerResponse = self.proximityGetCache[k] {
                    if let action:JSON? = response.action {
                        let beacon_id = self.getUUIDFromAction(action)
                        
                        if let place_id = action?["info"]["place_id"].asString {
                            if place_id == aPlaceID {
                                if alreadyReturned.indexForKey(beacon_id) == nil {
                                    alreadyReturned[beacon_id] = beacon_id
                                    result.append(action!)
                                }
                            }
                        }
                    }
                }
            }
        }
        
        return result
    }
    
    
    public func getUUIDFromAction(action: JSON?) -> String {
        if let inner_ids = action?["info"]["beacons"].asArray {
            for inner_id in inner_ids {
                if let beacon_id:String? = inner_id["beacon_id"].asString {
                    return beacon_id!
                }
            }
        }
        return ""
    }
    
    
    public func makeFullUUID(beacon:CLBeacon?) -> String {
        var fullUUID:String = ""
        
        if let uuid = beacon?.proximityUUID.UUIDString {
            if let major = beacon?.major {
                if var minor = beacon?.minor {
                    fullUUID = "\(uuid)_\(major)_\(minor)"
                }
            }
        }
        
        return fullUUID
    }
    
    
    public func isPreIOS8() -> Bool {
        if NSFoundationVersionNumber <= NSFoundationVersionNumber_iOS_7_1 {
            return true
        } else {
            return false
        }
    }
    
    
    public func sendNearestBeaconAsLocalNotification() {
        var newNotification:String = self.makeFullUUID(self.nearestBeaconWithActions!) + " " + self.getBeaconProximityAsString(self.nearestBeaconWithActions!)
        
        if newNotification != self.lastBeaconNotification {
            self.sendLocalNotification(newNotification)
            self.lastBeaconNotification = newNotification
        }
    }
    
    
    public func sendLocalNotification(message:String) {
        var localNotification:UILocalNotification = UILocalNotification()
        localNotification.fireDate = NSDate(timeIntervalSinceNow: 10)
        localNotification.alertBody = message;
        localNotification.timeZone = NSTimeZone.defaultTimeZone()
        UIApplication.sharedApplication().scheduleLocalNotification(localNotification)
    }
    
    
    public func saveUserName(name: String) {
        NSUserDefaults.standardUserDefaults().setObject(name, forKey:"MTOneSDKUsername")
        NSUserDefaults.standardUserDefaults().synchronize()
    }
    
    
    public func getUserName() -> String {
        var myUserName: String? = NSUserDefaults.standardUserDefaults().objectForKey("MTOneSDKUsername") as? String
        
        if myUserName == nil {
            myUserName = ""
        }
        
        return myUserName!
    }
    
    
    public func saveToken() {
        var myToken:String? = self.token
        if myToken == nil {
            myToken = ""
        }
        
        NSUserDefaults.standardUserDefaults().setObject(myToken, forKey:"MTOneSDKToken")
        NSUserDefaults.standardUserDefaults().synchronize()
        
    }
    
    
    public func getToken() -> String {
        var myToken: String? = NSUserDefaults.standardUserDefaults().objectForKey("MTOneSDKToken") as? String
        if myToken == nil {
            myToken = ""
        }
        
        self.token = myToken
        
        return myToken!
    }
    
    
    public func isValidToken(completion: (response: Bool)-> Void) {
        // Try to get the saved token if it exists
        if self.token == nil || self.token == "" {
            self.getToken()
        }
        
        if self.token == "" {
            return completion(response: false)
        }
        
        self.authGNDN { (response) -> Void in
            return completion(response: response.success)
        }
    }
    
    
    public func locationManager(manager: CLLocationManager!, didDetermineState state: CLRegionState, forRegion region: CLRegion!) {
        if (state == .Inside) {
            manager.startRangingBeaconsInRegion(region as CLBeaconRegion)
        } else {
            //Stop Ranging
            manager.stopMonitoringForRegion(region)
        }
    }
    
    
    public func isValidEmail(testStr:String) -> Bool {
        let emailRegEx = "[A-Z0-9a-z._%+-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,4}"
        
        var emailTest = NSPredicate(format:"SELF MATCHES %@", emailRegEx)
        let result = emailTest!.evaluateWithObject(testStr)
        return result
    }
    
    
    public func getAddressFromLocation(completion: (address:MTOneAddress?) -> Void) {
        if self.lastLocation != nil {
            CLGeocoder().reverseGeocodeLocation(self.lastLocation, completionHandler:
                {(placemarks, error) in
                    if (error != nil) {
                        println("reverse geodcode fail: \(error.localizedDescription)")
                        completion(address: nil)
                    }
                    
                    let pm = placemarks as [CLPlacemark]
                    if pm.count > 0 {
                        var address: MTOneAddress = MTOneAddress()
                        let placemark = placemarks[0] as? CLPlacemark
                        
                        if placemark!.subThoroughfare != nil {
                            address.number = placemark!.subThoroughfare
                        }
                        
                        if placemark!.thoroughfare != nil {
                            address.street = placemark!.thoroughfare
                        }
                        
                        // By default, this will be nil so we test for it
                        let cityObject:AnyObject? = placemark!.addressDictionary[kABPersonAddressCityKey]
                        if cityObject != nil {
                            address.city = cityObject as String
                        }
                        
                        let stateObject:AnyObject? = placemark!.addressDictionary[kABPersonAddressStateKey]
                        if stateObject != nil {
                            address.state = stateObject as String
                        }
                        
                        let postalObject:AnyObject? = placemark!.addressDictionary[kABPersonAddressZIPKey]
                        if postalObject != nil {
                            address.postal = postalObject as String
                        }
                        
                        address.type = "home"
                        
                        completion(address: address)
                    } else {
                        completion(address: nil)
                    }
            })
        } else {
            completion(address: nil)
        }
    }
    
    
    // Takes the current location and returns a placemark for it.
    public func getPlacemarkFromLocation(completion: (placemark: CLPlacemark?) -> Void) {
        var placemark:CLPlacemark? = nil
        
        if self.lastLocation != nil {
            CLGeocoder().reverseGeocodeLocation(self.lastLocation, completionHandler:
                {(placemarks, error) in
                    if (error != nil) {
                        println("reverse geodcode fail: \(error.localizedDescription)")
                        completion(placemark: nil)
                    }
                    
                    let pm = placemarks as [CLPlacemark]
                    if pm.count > 0 {
                        placemark = placemarks[0] as? CLPlacemark
                        
                        completion(placemark: placemark)
                    } else {
                        completion(placemark: nil)
                    }
            })
        } else {
            completion(placemark: nil)
        }
    }
    
    
    public func getLocationAsPoint() -> MTOnePoint {
        var point:MTOnePoint = MTOnePoint()
        
        if self.lastLocation != nil {
            point.lat = self.lastLocation!.coordinate.latitude
            point.lng = self.lastLocation!.coordinate.longitude
        }
        
        return point
    }
    
    /*
    public func voiceTTS(utterance: String, local:Bool = true) {
    if local == true {
    var speechSynthesizer:AVSpeechSynthesizer = AVSpeechSynthesizer()
    var speechUtterance:AVSpeechUtterance = AVSpeechUtterance(string:utterance)
    speechUtterance.rate = AVSpeechUtteranceMaximumSpeechRate / 5
    speechUtterance.voice = AVSpeechSynthesisVoice(language: "en-GB")
    
    speechSynthesizer.speakUtterance(speechUtterance)
    } else {
    // Call the server and then speak it here.
    }
    }
    */
    
    
    // Changes the resolution of GPS based on background mode
    public func applicationDidEnterBackground() {
        if self.usingGPS {
            locationManager.desiredAccuracy = kCLLocationAccuracyHundredMeters
            locationManager.distanceFilter = kCLDistanceFilterNone
            locationManager.pausesLocationUpdatesAutomatically = false
            locationManager.startUpdatingLocation()
        }
    }
    
    
    // Changes the resolution of GPS based on foreground
    public func applicationDidBecomeActive() {
        if self.usingGPS {
            locationManager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters
            locationManager.distanceFilter = kCLDistanceFilterNone
            locationManager.pausesLocationUpdatesAutomatically = false
            locationManager.startUpdatingLocation()
        }
    }
    
    
    private func boolToString(value: Bool) -> String {
        if value {
            return "true"
        } else {
            return "false"
        }
    }
    
    
    private func doubleToString(value: Double) -> String {
        return String(format: "%.6f", value)
    }
}


extension UIImage {
    // Rsizes an image that matches the desired dimensions
    // Our target is iPad air sized images. 2048x1536
    public class func smartResize(image: UIImage, width: Int, height: Int) -> UIImage {
        var aWidth:Int = width
        var aHeight:Int = height
        
        if image.size.width < 200 && image.size.height < 200 {
            return image
        }
        
        if width > height {
            if image.size.width < image.size.height {
                aWidth = height
                aHeight = width
            }
        } else {
            if image.size.height < image.size.width {
                aWidth = height
                aHeight = width
            }
        }
        
        
        return UIImage.resizeImage(image, width: aWidth, height: aHeight)
    }
    
    
    public func resizeToBoundingSquare(#boundingSquareSideLength : CGFloat) -> UIImage {
        let imgScale = self.size.width > self.size.height ? boundingSquareSideLength / self.size.width : boundingSquareSideLength / self.size.height
        let newWidth = self.size.width * imgScale
        let newHeight = self.size.height * imgScale
        let newSize = CGSize(width: newWidth, height: newHeight)
        
        UIGraphicsBeginImageContext(newSize)
        
        self.drawInRect(CGRect(x: 0, y: 0, width: newWidth, height: newHeight))
        
        let resizedImage = UIGraphicsGetImageFromCurrentImageContext()
        
        UIGraphicsEndImageContext();
        
        return resizedImage
    }
    
    
    public class func resizeImage(image: UIImage, width: Int, height: Int) -> UIImage {
        var newSize:CGSize = CGSize(width: width, height: height)
        let rect = CGRectMake(0,0, newSize.width, newSize.height)
        UIGraphicsBeginImageContextWithOptions(newSize, false, 1.0)
        
        // image is a variable of type UIImage
        image.drawInRect(rect)
        
        let newImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        
        return newImage
    }
}


extension NSData {
    func hexString() -> NSString {
        var str = NSMutableString()
        let bytes = UnsafeBufferPointer<UInt8>(start: UnsafePointer(self.bytes), count:self.length)
        for byte in bytes {
            str.appendFormat("%02hhx", byte)
        }
        return str
    }
}


let MTOneSharedInstance = MTOneSDK()
let MTOneSharedLogger = MTOneLogger()



